package org.objectweb.fractal.spoonlet.binding;

import static org.objectweb.fractal.spoonlet.binding.BindingHelper.BC;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasController;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasFieldAnnotation;
import static spoon.template.Substitution.insertAll;
import static spoon.template.Substitution.substituteMethodBody;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.fraclet.annotations.Requires;

import spoon.processing.AbstractAnnotationProcessor;
import spoon.reflect.code.CtStatement;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtField;
import spoon.reflect.reference.CtTypeReference;
import spoon.template.Template;

/**
 * Abstract binding processor.
 * 
 * @author Romain Rouvoy (Initial Developer)
 * @version $Revision$
 * @param <A> Type of annotation processed by the binding processor.
 */
public abstract class AbstractBindingProcessor<A extends Annotation> extends
        AbstractAnnotationProcessor<A, CtField<?>> {
    protected static final String[] KEYWORDS = { "listFc", "bindFc",
            "unbindFc", "lookupFc" };

    /* (non-Javadoc)
     * @see spoon.processing.AbstractAnnotationProcessor#inferConsumedAnnotationType()
     */
    @Override
    public boolean inferConsumedAnnotationType() {
        return false;
    }

    /* (non-Javadoc)
     * @see spoon.processing.AnnotationProcessor#process(java.lang.annotation.Annotation, spoon.reflect.declaration.CtElement)
     */
    public void process(A ann, CtField<?> fld) {
        CtClass<?> comp = fld.getParent(CtClass.class);
        getEnvironment().debugMessage(
                "[Fractal] Processing field " + comp.getQualifiedName() + "."
                        + fld.getSimpleName() + "...");
        insertMethodSignatures(comp, fld, ann);
        insertMethodBodies(comp, fld, ann);
    }

    /**
     * Inserts the method signatures associated to the processed field.
     * 
     * @param comp the component processed.
     * @param fld the field processed.
     * @param ann the annotation processed.
     */
    protected void insertMethodSignatures(CtClass<?> comp, CtField<?> fld, A ann) {
        CtTypeReference<BindingController> bc = getFactory().Interface()
                .createReference(BC);
        if (hasController(comp.getReference(), bc, false))
            return; // Class already implements BindingController (BC)
        CtTypeReference<?> ref = comp.getSuperclass();
        if (ref == null) // No parent class
            createMethodSignatures(comp);
        else if (hasController(ref, bc, true) // Parent class implements BC
                || hasFieldAnnotation(ref, Requires.class, true))
            insertAll(comp, new BindingExtensionTemplate());
        else
            createMethodSignatures(comp);
    }
    
    protected abstract void createMethodSignatures(CtClass<?> comp);

    /**
     * Insert a new body in the methods.
     * 
     * @param c reference to the class
     * @param f reference to the field
     * @param a reference to the annotation
     */
    protected abstract void insertMethodBodies(CtClass<?> c, CtField<?> f, A a);

    /**
     * Insert a piece of code in at the beginning of a method.
     * 
     * @param t template used for generating the piece of code to insert.
     * @param c target class for the insertion
     * @param m target method for the insertion
     * @param a method parameters
     * @return
     */
    protected void insertBegin(Template t, CtClass<?> c, String m,
            Class<?>... a) {
        List<CtTypeReference<?>> types = new ArrayList<CtTypeReference<?>>();
        for (Class<?> cls : a)
            types.add(getFactory().Type().createReference(cls));
        CtTypeReference<?>[] args = types.toArray(new CtTypeReference<?>[0]);
        CtStatement code = substituteMethodBody(c, t, m, args).getStatements()
                .get(0);
        c.getMethod(m, args).getBody().insertBegin(code);
    }
}